package tools.descartes.dql.connector.kieker;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import org.eclipse.emf.ecore.util.EcoreUtil;

import edu.kit.ipd.descartes.solving.probes.AvgThroughputProbe;
import tools.descartes.dql.connector.AbstractOSGiQueryConnector;
import tools.descartes.dql.connector.PerformanceMetricsQueryConnector;
import tools.descartes.dql.connector.QueryProgressListener;
import tools.descartes.dql.connector.kieker.structures.Resource;
import tools.descartes.dql.connector.kieker.structures.Trace;
import tools.descartes.dql.connector.kieker.structures.identifier.ResourceIdentifier;
import tools.descartes.dql.connector.kieker.structures.records.ResourceRecord;
import tools.descartes.dql.models.mapping.domain.DecimalResult;
import tools.descartes.dql.models.mapping.domain.DomainFactory;
import tools.descartes.dql.models.mapping.domain.Entity;
import tools.descartes.dql.models.mapping.domain.Probe;
import tools.descartes.dql.models.mapping.mapping.EntityMapping;

public class PerformanceMetricsQueryConnectorImpl extends
		AbstractOSGiQueryConnector implements PerformanceMetricsQueryConnector {

	private static String STANDART_BAT_PROPERTIES_KEY = "kieker.standartBat";
	private static String MAX_TIMEOUT_KEY = "kieker.maxRunTimeInSeconds";
	private static String KIEKER_CONFIG_KEY = "kieker.monitoring.default.properties";

	private final DomainFactory domainFactory = DomainFactory.eINSTANCE;


	@Override
	public EntityMapping requestMetrics(final EntityMapping request) {
		System.out.println("Starting Kieker Controller now" + request.toString());

		final EntityMapping response = EcoreUtil.copy(request);
		final String propertiesLocation = request.getModelLocation();

		// Properties laden
		Properties prop = getProperties(propertiesLocation);
		String standardBatLocation = (String) prop.get(STANDART_BAT_PROPERTIES_KEY);
		String standardBatPath = standardBatLocation.substring(0,
				standardBatLocation.lastIndexOf(File.separator) + File.separator.length());
		String standardBatName = standardBatLocation.substring(
				standardBatLocation.lastIndexOf(File.separator) + File.separator.length());
		//TODO implement wait for end
		// TODO implement warning if no int
		int timeOut = Integer.valueOf(prop.getProperty(MAX_TIMEOUT_KEY));
		String kiekerConfigLocation = prop.getProperty(KIEKER_CONFIG_KEY);
		Properties kiekerProperties = loadKiekerProperties(kiekerConfigLocation);
		String newOutputPath = getNewOutPutPath(standardBatPath);
		kiekerProperties.setProperty("kieker.monitoring.writer.filesystem.AsyncFsWriter.customStoragePath",
				newOutputPath);
		try {
			kiekerProperties.store(new FileOutputStream(kiekerConfigLocation), "");
		} catch (FileNotFoundException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
		
			
		// Kieker starten
		// TODO kieker.monitoring.default.properties verndern/erstellen
		try {
			Runtime.getRuntime().exec("cmd /c \"cd " + standardBatPath + " &&  start " + standardBatName + "\"");
		} catch (IOException e) {
			e.printStackTrace();
			throw new IllegalStateException("Could not execute .bat file");
		}

		// warten bis timeout
		try {
			Thread.sleep(timeOut * 1000);
			//TODO run directly to avoid having to kill java.exe
			Runtime.getRuntime().exec("taskkill /f /im java.exe");
			Runtime.getRuntime().exec("taskkill /f /im cmd.exe");
		} catch (InterruptedException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}

		// Kieker ergebnisse einlesen


		// Kieker ergebnisse filtern

		// Start KiekerController with custom config (currently only location of
		// traceFiles)
		// KiekerController c = new KiekerController(propertiesLocation +
		// "//controller.cfg");

		// Fill the traces according to the applied filters
		// c.fillTrace();

		// Get the result trace
		// Trace result = c.getTrace();

		// Process the trace back to a DQL result
		// processKiekerSolverResults(response, result);
		return response;
	}

	private String getNewOutPutPath(String standardBatPath) {
		standardBatPath += File.separator + "logs";
		File logFolder = new File(standardBatPath);
		if (!logFolder.exists()) {
			logFolder.mkdir();
			logFolder.mkdirs();
		}
		int i = 1;
		File subLogFolder = null;
		do {
			subLogFolder = new File(logFolder, "run" + i);
			i++;
		} while (subLogFolder.exists());
		subLogFolder.mkdir();
		subLogFolder.mkdirs();
		return subLogFolder.getAbsolutePath();
	}

	/* Modifies response to include the results of Kieker */
	private void processKiekerSolverResults(EntityMapping response, Trace kiekerResults) {

		// Map containing the resulting records sorted by resources
		final HashMap<String, List<ResourceRecord>> kiekerResultMap = new HashMap<String, List<ResourceRecord>>(
				kiekerResults.getResourceIdentifierToResourceMap().size());

		// List of Records for one resource, value for kiekerResultMap
		List<ResourceRecord> records = kiekerResults.getResources();

		// kieker resource id to resource mapping, help for identifying the
		// records
		Map<ResourceIdentifier, Resource> resourceMap = kiekerResults.getResourceIdentifierToResourceMap();

		// for every retrieved record get the original dql identifier, if no
		// match found but still in map, throw away)
		// TODO: should be avoided by resource filter
		for (ResourceRecord kiekerProbe : records) {
			if (resourceMap.get(kiekerProbe.getIdentifier()) != null) {
				String identifier = resourceMap.get(kiekerProbe.getIdentifier()).getName();
				List<ResourceRecord> kiekerProbes = kiekerResultMap.get(identifier);
				if (null == kiekerProbes) {
					kiekerProbes = new ArrayList<ResourceRecord>();
					kiekerProbes.add(kiekerProbe);
					kiekerResultMap.put(identifier, kiekerProbes);

				} else {
					kiekerProbes.add(kiekerProbe);
					kiekerResultMap.put(identifier, kiekerProbes);

				}
			}

		}
		
		final ArrayList<Entity> entities = new ArrayList<Entity>();
		entities.addAll(response.getResources());
		entities.addAll(response.getServices());
		
		// browse through each entity and its probes 
		//   and try to replace the probe with a result
		for (Entity ent : entities) {
			for (Probe dqlProbe : ent.getProbes()) {
				System.out.println("ORIGIN ID: " + ent.getIdentifier().toString());
				List<ResourceRecord> kiekerProbes = kiekerResultMap.get( ent.getIdentifier().toString());
				String value = null;
				if (kiekerProbes != null) {
					if (DMLHelper.METRIC_RESPONSE_TIME_AVG.equalsIgnoreCase(dqlProbe.getMetricName())) {
						for (ResourceRecord kiekerProbe : kiekerProbes) {
							value = kiekerProbe.toString();
							break;
						}
					} else if (DMLHelper.METRIC_UTILIZATION.equalsIgnoreCase(dqlProbe.getMetricName())) {
						for (ResourceRecord kiekerProbe : kiekerProbes) {
							if (true) {
								value = String.valueOf(kiekerProbe.getContent().get(2));
								break;
							}
						}
					} else if (DMLHelper.METRIC_THROUGHPUT.equalsIgnoreCase(dqlProbe.getMetricName())) {
						for (ResourceRecord dmlProbe : kiekerProbes) {
							if (dmlProbe instanceof AvgThroughputProbe) {
								value = ((AvgThroughputProbe) dmlProbe).getValue();
								break;
							}
						}
					} else {
						// do nothing
					}
				}
				
				if (null != value) {
					DecimalResult res = domainFactory.createDecimalResult();
					res.setValid(true);
					try {
						BigDecimal val = new BigDecimal(value);
						res.setValue(val);
					} catch (NumberFormatException exc) {
						res.setValid(false);
					}	
					dqlProbe.getStatTypes().get(0).setResult(res);
				}
			}
		}
	}

	private Properties loadKiekerProperties(final String propertiesLocation) {
		final Properties properties = new Properties();
		File file = new File(propertiesLocation);

		if (file == null || !file.exists()) {
			throw new IllegalStateException(
					"Could not find referenced .properties file");
		}

		try {
			properties.load(new FileInputStream(file));
		} catch (final IOException e) {
			throw new IllegalStateException(
					"Could not load properties instance");
		}
		return properties;
	}

	private Properties getProperties(final String propertiesLocation) {
		final Properties properties = new Properties();
		File file = new File(propertiesLocation);

		if (file == null || !file.exists()) {
			throw new IllegalStateException(
					"Specified file in USING not accessible, unable to execute the query further. Aborting.");
		}

		try {
			properties.load(new FileInputStream(file));
		} catch (final IOException e) {
			throw new IllegalStateException(
					"Could not load properties instance from specified model location in USING clause");
		}
		// Add absolutePath for the bat file
		String propertiesFolder = new File(propertiesLocation).getParentFile().getAbsolutePath();

		properties.setProperty(STANDART_BAT_PROPERTIES_KEY,
				propertiesFolder + File.separator + properties.get(STANDART_BAT_PROPERTIES_KEY));
		properties.setProperty(KIEKER_CONFIG_KEY,
				propertiesFolder + File.separator + properties.get(KIEKER_CONFIG_KEY));
		return properties;
	}

	@Override
	public void reset() {
		// Nothing to do here
	}

	@Override
	public void addQueryProgressListener(QueryProgressListener arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public String getConstraint() {
		return null;
	}

	@Override
	public String getDoFConstraint() {
		return null;
	}

	@Override
	public String getDoFExplorationStrategy() {
		return null;
	}

	@Override
	public Properties getDoFExplorationStrategyProperties() {
		return null;
	}

	@Override
	public List<String> getSupportedConstraintNames() {
		return null;
	}

	@Override
	public List<String> getSupportedDoFConstraintNames() {
		return null;
	}

	@Override
	public List<String> getSupportedExplorationStrategyNames() {
		return null;
	}

	@Override
	public List<String> getSupportedTimeUnits() {
		return null;
	}

	@Override
	public List<EntityMapping> requestDoFMetrics(final EntityMapping request) {
		return null;
	}
	
	@Override
	public void setConstraint(final String constraint) {
	}

	@Override
	public void setDoFConstraint(final String doFConstraint) {
	}

	@Override
	public void setDoFExplorationStrategy(final String strategy,
			final Properties properties) {
	}

	@Override
	public void setObservationPeriod(final Date start, final Date end,
			final long sampleSize) {
	}

	@Override
	public void setObservationPeriod(final long amount, final String unit,
			final long sampleSize) {
	}

	@Override
	public void setObservationReference(final String type,
			final String identifier, final long sampleSize) {
	}

	@Override
	public boolean supportsConstraintRequests() {
		return false;
	}

	@Override
	public boolean supportsDoFRequests() {
		return false;
	}

	@Override
	public boolean supportsMetricRequests() {
		return true;
	}

	@Override
	public boolean supportsObservationPeriods() {
		return false;
	}
}
